
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "extGerberDrillFileReader.h"

namespace ext
{

// ---------------------------------------------------------------------------------
//  GerberDrillFileReader implementation

GerberDrillFileReader::GerberDrillFileReader ()
  : GerberFileReader ()
{
  // .. nothing yet ..
}

GerberDrillFileReader::~GerberDrillFileReader ()
{
  // .. nothing yet ..
}

bool 
GerberDrillFileReader::does_accept ()
{
  for (int i = 0; i < 100; ++i) {
    tl::Extractor ex (get_block ().c_str ());
    if (ex.test ("M71")) {
      return true;
    } else if (ex.test ("M48")) {
      return true;
    } else if (ex.test (";")) {
      if (ex.test ("Holesize")) {
        return true;
      } else if (ex.test ("T")) {
        return true;
      }
    }
  }

  return false;
}

void
GerberDrillFileReader::do_read ()
{
  m_buffer.clear ();

  //  Initialize reader:
  bool relative = false;
  double x = 0.0, y = 0.0;
  double xoff = 0.0, yoff = 0.0;
  double current_diameter = 0.0;

  std::map <int, double> tools;

  stream ().skip ();

  //  Actually read:
  while (! stream ().at_end ()) {

    tl::Extractor ex (get_block ().c_str ());
    char c = *ex.skip ();

    if (ex.test ("%")) {

      //  blocks starting with a '%' are ignored.

    } else if (ex.test (";")) {

      if (ex.test ("Holesize")) {

        double d = 0.0;
        ex.read (d);
        ex.test ("=");
        ex.read (current_diameter);
        current_diameter *= unit ();

      } else if (ex.test ("size")) {

        ex.test (":");
        ex.read (current_diameter);
        current_diameter *= unit ();

      } else if (ex.test ("T")) {

        int t = 0;
        if (ex.try_read (t)) {

          if (ex.test ("Holesize")) {

            double d = 0.0;
            ex.read (d);
            ex.test ("=");
            d = 0.0;
            ex.read (d);
            d *= unit ();

            tools.insert (std::make_pair (t, d));

          } else if (ex.test ("size")) {

            ex.test (":");
            double d = 0.0;
            ex.read (d);
            d *= unit ();

            tools.insert (std::make_pair (t, d));

          }

        }

      } else if (ex.test ("FILE_FORMAT")) {

        ex.test ("=");

        int l = -1, t = -1;
        ex.read (l);
        ex.expect (":");
        ex.read (t);

        set_format (l, t);

      }

    } else if (ex.test ("FMAT")) {

      //  TODO: what to do here?

    } else if (ex.test ("ICI")) {

      //  TODO: what to do here?

    } else if (ex.test ("INCH")) {

      set_unit (25400.0);

      if (ex.test (",")) {
        if (ex.test ("LZ")) {
          set_format (false); // omit trailing zeroes
        } else if (ex.test ("TZ")) {
          set_format (true); // omit leading zeroes
        }
      }

      ex.expect_end ();

    } else if (ex.test ("METRIC")) {

      set_unit (1000.0);

      if (ex.test (",")) {
        if (ex.test ("LZ")) {
          set_format (false); // omit trailing zeroes
        } else if (ex.test ("TZ")) {
          set_format (true); // omit leading zeroes
        }
      }

      ex.expect_end ();

    } else if (ex.test ("M")) {

      int mcode = 0;
      ex.read (mcode);
      ex.expect_end ();

      // @@@ process M-code

    } else if (ex.test ("T")) {

      int tcode = 0;
      ex.read (tcode);

      c = *ex.skip ();

      if (c == 'F' || c == 'S' || c == 'C') {

        while (true) {
          double d = 0.0;
          if (ex.test ("F") || ex.test ("S")) {
            ex.read (d);
          } else if (ex.test ("C")) {
            ex.read (d);
            d *= unit ();
            tools.insert (std::make_pair (tcode, d));
            current_diameter = d;
          } else {
            break;
          }
        }

      } else if (ex.test ("size")) {

        double d = 0.0;

        ex.test (":");
        ex.read (d);
        d *= unit ();

        tools.insert (std::make_pair (tcode, d));
        current_diameter = d;

      } else {

        ex.expect_end ();

        if (tools.find (tcode) == tools.end ()) {
          if (tcode == 0) {
            //  some file formats indicate "no tool" with this code ..
            current_diameter = 0.0;
          } else {
            throw tl::Exception (tl::translate ("Undefined tool code %d").c_str (), tcode);
          }
        } else {
          current_diameter = tools [tcode];
        }

      }

    } else if (ex.test ("R")) {

      int n = 0;
      ex.read (n);

      double dx = 0.0, dy = 0.0;

      if (ex.test ("X")) {
        dx = read_coord (ex);
      }

      if (ex.test ("Y")) {
        dy = read_coord (ex);
      }

      for (int i = 1; i <= n; ++i) {
        produce_circle (x + i * dx, y + i * dy, current_diameter * 0.5);
      }
      
      ex.expect_end ();

    } else if (c == 'X' || c == 'Y' || c == 'G') {

      bool has_coord = false;

      int gcode = 0;

      while (! ex.at_end ()) {

        if (ex.test ("G")) {

          ex.read (gcode);

          if (gcode == 90) {
            relative = false;
          } else if (gcode == 91) {
            relative = true;
          } 

        } else if (ex.test ("X")) {

          double d = read_coord (ex);
          if (relative) {
            x += d;
          } else {
            x = d;
          }

          has_coord = true;

        } else if (ex.test ("Y")) {

          double d = read_coord (ex);
          if (relative) {
            y += d;
          } else {
            y = d;
          }

          has_coord = true;

        } else {
          break;
        }

      }

      if (has_coord) {
        if (gcode == 93 || gcode == 92) {
          xoff = x; yoff = y;
        } else {
          produce_circle (x + xoff, y + yoff, current_diameter * 0.5);
        }
      }

      if (! ex.at_end ()) {
        warn (tl::sprintf (tl::translate ("Part of line ignored: %s").c_str (), ex.skip ()));
      }

    } else if (c != 0) {
      warn (tl::translate ("Statement ignored"));
    }

    stream ().skip ();

  }
}

const std::string &
GerberDrillFileReader::get_block ()
{
  progress_checkpoint ();

  m_buffer.clear ();

  char c;
  while (! stream ().at_end ()) {
    c = stream ().get_char ();
    if (c == '\n' || c == '\r') {
      break;
    }
    m_buffer += c;
  }

  c = stream ().peek_char ();
  if (c == '\n' || c == '\r') {
    stream ().get_char ();
  }

  return m_buffer;
}

void
GerberDrillFileReader::produce_circle (double cx, double cy, double r)
{
  std::vector <db::DPoint> points;

  int n_circle = get_circle_points ();

  //  adjust the radius so we get a outer approximation of the circle:
  //  r *= 1.0 / cos (M_PI / double (n_circle));

  for (int i = 0; i < n_circle; ++i) {
    double a = M_PI * 2.0 * ((double (i) + 0.5) / double (n_circle));
    points.push_back (db::DPoint (cx + r * cos (a), cy + r * sin (a)));
  }

  db::DPolygon p;
  p.assign_hull (points.begin (), points.end ());

  produce_polygon (p, false);
}

}

