
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "extXORToolDialog.h"
#include "rdb.h"
#include "dbEdgeProcessor.h"
#include "dbLayoutUtils.h"
#include "dbClip.h"
#include "tlTimer.h"
#include "tlProgress.h"
#include "layCellView.h"
#include "layLayoutView.h"
#include "layExceptions.h"

#include "XORToolDialog.h"

#include <stdio.h>

namespace ext
{

XORToolDialog::XORToolDialog (QWidget *parent)
{
  mp_ui = new Ui::XORToolDialog ();
  mp_ui->setupUi (this);

  connect (mp_ui->input_layers_cbx, SIGNAL (currentIndexChanged (int)), this, SLOT (input_changed (int)));
  connect (mp_ui->output_cbx, SIGNAL (currentIndexChanged (int)), this, SLOT (output_changed (int)));

  input_changed (0);
  output_changed (0);
}

XORToolDialog::~XORToolDialog ()
{
  delete mp_ui;
  mp_ui = 0;
}

int 
XORToolDialog::exec (lay::LayoutView *view)
{
  mp_view = view;

  if (view != mp_ui->layouta->layout_view () || view != mp_ui->layoutb->layout_view ()) {

    mp_ui->layouta->set_layout_view (view);
    mp_ui->layoutb->set_layout_view (view);

    if (view->cellviews () >= 2) {
      mp_ui->layouta->set_current_cv_index (0);
      mp_ui->layoutb->set_current_cv_index (1);
    }

  } else {
    //  force update of the layer list
    //  TODO: the controls should register a listener for the view so this activity is not necessary:
    mp_ui->layouta->set_layout_view (view);
    mp_ui->layoutb->set_layout_view (view);
  }

  int ret = QDialog::exec ();

  if (ret) {
    run_xor ();
  }

  mp_view = 0;
  return ret;
}

void 
XORToolDialog::accept ()
{
BEGIN_PROTECTED

  bool axorb = mp_ui->axorb_cb->isChecked ();
  bool anotb = mp_ui->anotb_cb->isChecked ();
  bool bnota = mp_ui->bnota_cb->isChecked ();
  if (axorb + anotb + bnota == 0) {
    throw tl::Exception (tl::translate ("No mode selected"));  
  }

  int cv_index_a = mp_ui->layouta->current_cv_index ();
  int cv_index_b = mp_ui->layoutb->current_cv_index ();

  const lay::CellView &cva = mp_view->cellview (cv_index_a);
  const lay::CellView &cvb = mp_view->cellview (cv_index_b);

  if (&cva->layout () == &cvb->layout () && cva.cell_index () == cvb.cell_index ()) {
    throw tl::Exception (tl::translate ("Trying to perform an XOR between identical layouts"));  
  }

  // TODO: change this!
  if (fabs (cva->layout ().dbu () - cvb->layout ().dbu ()) > 1e-6) {
    throw tl::Exception (tl::translate ("Both layouts must have the same database unit currently"));  
  }

  {
    tl::Extractor ex (mp_ui->tolerances->text ().ascii ());
    while (! ex.at_end ()) {
      double t;
      if (! ex.try_read (t) || t < -1e-6) {
        break;
      }
      ex.test (",");
    }
  }

  {
    tl::Extractor ex (mp_ui->tiling->text ().ascii ());
    double t = 0.0;
    if (ex.try_read (t)) {
      if (t < 0.001) {
        throw tl::Exception (tl::translate ("Invalid tile size (invalid text or negative)"));  
      }
    }
  }
  
  QDialog::accept ();

END_PROTECTED
}

void
XORToolDialog::input_changed (int index)
{
  // .. nothing yet ..
}

void
XORToolDialog::output_changed (int index)
{
  bool enabled = (index == 2 || index == 3);
  mp_ui->layer_offset_lbl->setEnabled (enabled);
  mp_ui->layer_offset_le->setEnabled (enabled);
}

void 
XORToolDialog::run_xor ()
{
  enum input_mode_t { IMAll, IMVisible, IMSpecific };
  input_mode_t input_mode = (input_mode_t) mp_ui->input_layers_cbx->currentIndex ();

  enum output_mode_t { OMMarkerDatabase, OMNewLayout, OMNewLayersA, OMNewLayersB };
  output_mode_t output_mode = (output_mode_t) mp_ui->output_cbx->currentIndex ();

  db::LayerOffset layer_offset;
  if (output_mode == OMNewLayersA || output_mode == OMNewLayersB) {
    std::string lo = mp_ui->layer_offset_le->text ().ascii ();
    tl::Extractor ex (lo.c_str ());
    layer_offset.read (ex);
  }

  bool axorb = mp_ui->axorb_cb->isChecked ();
  bool anotb = mp_ui->anotb_cb->isChecked ();
  bool bnota = mp_ui->bnota_cb->isChecked ();

  bool summarize = mp_ui->summarize_cb->isChecked ();

  int cv_index_a = mp_ui->layouta->current_cv_index ();
  int cv_index_b = mp_ui->layoutb->current_cv_index ();

  lay::CellView cva = mp_view->cellview (cv_index_a);
  lay::CellView cvb = mp_view->cellview (cv_index_b);

  // TODO: change this!
  double dbu = cva->layout ().dbu ();

  std::map<db::LayerProperties, std::pair<int, int>, db::LPLogicalLessFunc> layers;

  for (db::Layout::layer_iterator la = cva->layout ().begin_layers (); la != cva->layout ().end_layers (); ++la) {
    layers.insert (std::make_pair (*(*la).second, std::make_pair (-1, -1))).first->second.first = int ((*la).first);
  }

  for (db::Layout::layer_iterator lb = cvb->layout ().begin_layers (); lb != cvb->layout ().end_layers (); ++lb) {
    layers.insert (std::make_pair (*(*lb).second, std::make_pair (-1, -1))).first->second.second = int ((*lb).first);
  }

  //  Keep only visible layers if requested. Treat invisible ones as empty.
  if (input_mode == IMVisible) {

    std::set<int> visible_layers_a;
    std::set<int> visible_layers_b;

    for (lay::LayerPropertiesConstIterator l = mp_view->begin_layers (); l != mp_view->end_layers (); ++l) {
      if (! l->has_children () && l->layer_index () >= 0 && l->cellview_index () == cv_index_a && l->visible (true)) {
        visible_layers_a.insert (int (l->layer_index ()));
      }
      if (! l->has_children () && l->layer_index () >= 0 && l->cellview_index () == cv_index_b && l->visible (true)) {
        visible_layers_b.insert (int (l->layer_index ()));
      }
    }

    for (std::map<db::LayerProperties, std::pair<int, int>, db::LPLogicalLessFunc>::iterator lm = layers.begin (); lm != layers.end (); ++lm) {
      if (lm->second.first >= 0 && visible_layers_a.find (lm->second.first) == visible_layers_a.end ()) {
        lm->second.first = -1;
      }
      if (lm->second.second >= 0 && visible_layers_b.find (lm->second.second) == visible_layers_b.end ()) {
        lm->second.second = -1;
      }
    }

    for (std::map<db::LayerProperties, std::pair<int, int>, db::LPLogicalLessFunc>::iterator lm = layers.begin (); lm != layers.end (); ) {
      std::map<db::LayerProperties, std::pair<int, int>, db::LPLogicalLessFunc>::iterator lm_next = lm;
      ++lm_next;
      if (lm->second.first < 0 && lm->second.second < 0) {
        layers.erase (lm);
      }
      lm = lm_next;
    }

  }

  std::vector <db::Coord> tolerances;

  {
    tl::Extractor ex (mp_ui->tolerances->text ().ascii ());
    while (! ex.at_end ()) {
      double t;
      if (! ex.try_read (t) || t < -1e-6) {
        break;
      }
      ex.test (",");
      tolerances.push_back (db::coord_traits <db::Coord>::rounded (t / dbu));
    }

    std::sort (tolerances.begin (), tolerances.end ());
    tolerances.erase (std::unique (tolerances.begin (), tolerances.end ()), tolerances.end ());

    if (tolerances.size () == 0) {
      tolerances.push_back (0);
    }
  }

  //  Create a map of new layers for original ones 
  std::map <db::LayerProperties, db::LayerProperties> new_layer_props;
  for (std::map<db::LayerProperties, std::pair<int, int>, db::LPLogicalLessFunc>::iterator lm = layers.begin (); lm != layers.end (); ++lm) {
    new_layer_props.insert (std::make_pair (lm->first, lm->first));
  }

  double tile_size = 0; // in micron units

  {
    tl::Extractor ex (mp_ui->tiling->text ().ascii ());
    double t = 0.0;
    if (ex.try_read (t)) {
      tile_size = t;
      if (tile_size < 1.0) {
        throw tl::Exception (tl::translate ("Invalid tile size (smaller than 1 micron or negative)"));  
      }
    }
  }
  
  std::string srca = cva->name () + ", Cell " + cva->layout ().cell_name (cva.cell_index ());
  std::string srcb = cvb->name () + ", Cell " + cvb->layout ().cell_name (cvb.cell_index ());

  //  Create the report database or identify the output layout
  rdb::Database *rdb = 0;
  rdb::Cell *rdb_cell = 0;
  int rdb_index = 0;

  int output_cv = -1;
  db::Layout *output_layout = 0;
  db::Cell *output_cell = 0;
  std::vector <unsigned int> output_layers;

  if (output_mode == OMMarkerDatabase) {

    rdb = new rdb::Database ();
    rdb->set_name ("XOR");
    rdb->set_top_cell_name (cva->layout ().cell_name (cva.cell_index ()));
    rdb_cell = rdb->create_cell (rdb->top_cell_name ());

    rdb_index = mp_view->add_rdb (rdb);

    rdb->set_description ("Comparison of '" + srca + "' vs. '" + srcb + "'");

  } else if (output_mode == OMNewLayout) {

    output_cv = mp_view->create_layout (true);
    output_layout = &mp_view->cellview (output_cv)->layout ();

    for (std::map <db::LayerProperties, db::LayerProperties>::const_iterator lp = new_layer_props.begin (); lp != new_layer_props.end (); ++lp) {
      output_layers.push_back (output_layout->insert_layer (lp->second));
      lay::LayerProperties lay_lp;
      lay_lp.set_source (lay::ParsedLayerSource (lp->second, output_cv));
      mp_view->init_layer_properties (lay_lp);
      mp_view->insert_layer (mp_view->end_layers (), lay_lp);
    }

  } else if (output_mode == OMNewLayersA) {

    output_cv = cv_index_a;
    output_layout = &cva->layout ();
    output_cell = cva.cell ();

  } else if (output_mode == OMNewLayersB) {

    output_cv = cv_index_b;
    output_layout = &cvb->layout ();
    output_cell = cvb.cell ();

  }

  //  Clear undo buffers if layout is created.
  if (output_layout) {
    mp_view->manager ()->clear ();
  }

  //  compute the tiles if required
  db::Box overall_box_a = cva.cell ()->bbox ();
  db::Box overall_box_b = cvb.cell ()->bbox ();

  db::Coord box_width_a = 0, box_height_a = 0;
  db::Coord box_width_b = 0, box_height_b = 0;

  db::DBox overall_box = (db::DBox (overall_box_a) * cva->layout ().dbu ()) + (db::DBox (overall_box_b) * cvb->layout ().dbu ());
  size_t ntiles_w = 1, ntiles_h = 1;
  if (tile_size > 0.0) {

    overall_box_a = db::Box (overall_box * (1.0 / cva->layout ().dbu ()));
    overall_box_b = db::Box (overall_box * (1.0 / cvb->layout ().dbu ()));

    ntiles_w = std::max (size_t (1), size_t (floor (overall_box.width () / tile_size + 0.5)));
    ntiles_h = std::max (size_t (1), size_t (floor (overall_box.height () / tile_size + 0.5)));

    box_width_a  = overall_box_a.width () / ntiles_w;
    box_height_a = overall_box_a.height () / ntiles_h;

    box_width_b  = overall_box_b.width () / ntiles_w;
    box_height_b = overall_box_b.height () / ntiles_h;

  }

  //  Enlarge the tiles by half the maximum tolerance
  db::Coord tile_enlargement = 0;
  for (std::vector <db::Coord>::iterator t = tolerances.begin (); t != tolerances.end (); ++t) {
    db::Coord enlargement = (*t + 1) / 2; // round up
    if (enlargement > tile_enlargement) {
      tile_enlargement = enlargement;
    }
  }

  bool has_tiles = (ntiles_w > 1 || ntiles_h > 1);

  db::ShapeProcessor sp (true);

  size_t nmodes = (anotb ? 1 : 0) + (bnota ? 1 : 0) + (axorb ? 1 : 0);
  size_t todo_count = ntiles_w * ntiles_h * nmodes * tolerances.size () * layers.size ();
  tl::Progress progress (tl::translate ("Running XOR tool"), todo_count, 1);

  for (int mode = 0; mode < 3; ++mode) {

    rdb::Category *cat = 0; 
    db::BooleanOp::BoolOp op;
    std::string op_name;
    std::string op_desc;

    if (mode == 0 && axorb) {
      op = db::BooleanOp::Xor;
      op_name = "XOR";
      op_desc = "XOR between '" + srca + "' (Layout A) and '" + srcb + "' (Layout B)";
    } else if (mode == 1 && anotb) {
      op = db::BooleanOp::ANotB;
      op_name = "ANOTB";
      op_desc = "Geometry in '" + srca + "' (Layout A) but not in '" + srcb + "' (Layout B)";
    } else if (mode == 2 && bnota) {
      op = db::BooleanOp::BNotA;
      op_name = "BNOTA";
      op_desc = "Geometry in '" + srca + "' (Layout B) but not in '" + srcb + "' (Layout A)";
    } else {
      continue;
    }

    if (output_mode == OMMarkerDatabase) {
      cat = rdb->create_category (op_name);
      cat->set_description (op_desc);
    } else if (output_mode == OMNewLayout) {
      output_cell = &output_layout->cell (output_layout->add_cell (op_name.c_str ()));
    }

    std::vector <rdb::Category *> sub_categories;
    std::vector <std::vector <rdb::Category *> > layer_categories;
    std::vector <db::Cell *> sub_cells;
    std::vector <std::vector <unsigned int> > sub_output_layers;

    if (output_mode == OMMarkerDatabase) {

      //  create the categories for database output

      if (tolerances.size () == 1) {

        sub_categories.push_back (cat);

      } else {

        for (std::vector <db::Coord>::const_iterator t = tolerances.begin (); t != tolerances.end (); ++t) {

          rdb::Category *subcat;
          subcat = rdb->create_category (cat, tl::sprintf ("Tol_%g", *t * dbu));
          subcat->set_description (tl::sprintf ("XOR tolerance (min width reported): %g um", *t * dbu));
          sub_categories.push_back (subcat);

        }

      }

      layer_categories.reserve (sub_categories.size ());
      for (size_t i = 0; i < sub_categories.size (); ++i) {

        layer_categories.push_back (std::vector <rdb::Category *> ());

        for (std::map<db::LayerProperties, std::pair<int, int>, db::LPLogicalLessFunc>::const_iterator l = layers.begin (); l != layers.end (); ++l) {

          rdb::Category *layercat = rdb->create_category (sub_categories [i], l->first.to_string ());
          layercat->set_description ("Results for layer " + l->first.to_string ());
          layer_categories.back ().push_back (layercat);

        }

      }

    } else if (output_mode == OMNewLayout) {

      if (tolerances.size () == 1) {

        sub_cells.push_back (output_cell);
        sub_output_layers.push_back (output_layers);

      } else {

        for (std::vector <db::Coord>::const_iterator t = tolerances.begin (); t != tolerances.end (); ++t) {

          sub_cells.push_back (&output_layout->cell (output_layout->add_cell (tl::sprintf ("%s_TOL_%g", op_name.c_str (), *t * dbu).c_str ())));
          output_cell->insert (db::CellInstArray (db::CellInst (sub_cells.back ()->cell_index ()), db::Trans ()));

          sub_output_layers.push_back (output_layers);

        }

      }

    } else if (output_mode == OMNewLayersA || output_mode == OMNewLayersB) {

      if (tolerances.size () == 1) {

        db::LayerOffset o = layer_offset;
        if (! o.is_named ()) {
          o.name = "*_" + op_desc; // "_XOR" postfix by default
        }

        for (std::map <db::LayerProperties, db::LayerProperties>::iterator lp = new_layer_props.begin (); lp != new_layer_props.end (); ++lp) {
          if (lp->first.is_named ()) {
            lp->second.name = lp->first.name;
          }
          lp->second += o;
        }

        output_layers.clear ();
        for (std::map <db::LayerProperties, db::LayerProperties>::iterator lp = new_layer_props.begin (); lp != new_layer_props.end (); ++lp) {
          output_layers.push_back (output_layout->insert_layer (lp->second));
          lay::LayerProperties lay_lp;
          lay_lp.set_source (lay::ParsedLayerSource (lp->second, output_cv));
          mp_view->init_layer_properties (lay_lp);
          mp_view->insert_layer (mp_view->end_layers (), lay_lp);
        }

        sub_cells.push_back (output_cell);
        sub_output_layers.push_back (output_layers);

      } else {

        for (std::vector <db::Coord>::const_iterator t = tolerances.begin (); t != tolerances.end (); ++t) {

          db::LayerOffset o = layer_offset;
          if (! o.is_named ()) {
            o.name = "*_" + op_desc + tl::sprintf ("_T%d", int (t - tolerances.begin ()) + 1); // "_XOR" postfix by default
          }

          for (std::map <db::LayerProperties, db::LayerProperties>::iterator lp = new_layer_props.begin (); lp != new_layer_props.end (); ++lp) {
            if (lp->first.is_named ()) {
              lp->second.name = lp->first.name;
            }
            lp->second += o;
          }

          output_layers.clear ();
          for (std::map <db::LayerProperties, db::LayerProperties>::iterator lp = new_layer_props.begin (); lp != new_layer_props.end (); ++lp) {
            output_layers.push_back (output_layout->insert_layer (lp->second));
            lay::LayerProperties lay_lp;
            lay_lp.set_source (lay::ParsedLayerSource (lp->second, output_cv));
            mp_view->init_layer_properties (lay_lp);
            mp_view->insert_layer (mp_view->end_layers (), lay_lp);
          }

          sub_cells.push_back (output_cell);
          sub_output_layers.push_back (output_layers);

        }

      }

    }

    tl::SelfTimer timer_tot (tl::verbosity () >= 11, "Total boolean time");

    //  start the XOR procedure
    for (size_t nw = 0; nw < ntiles_w; ++nw) {

      for (size_t nh = 0; nh < ntiles_h; ++nh) {

        //  prepare a layout for the results
        db::Layout xor_results;
        db::Cell &xor_results_cell = xor_results.cell (xor_results.add_cell ());

        db::Box clip_box (overall_box_a.left () + nw * box_width_a, 
                          overall_box_a.bottom () + nh * box_height_a,
                          (nw == ntiles_w - 1) ? overall_box_a.right () : overall_box_a.left () + (nw + 1) * box_width_a,
                          (nh == ntiles_h - 1) ? overall_box_a.top () : overall_box_a.bottom () + (nh + 1) * box_height_a);

        db::Box region_a (overall_box_a.left () + nw * box_width_a, 
                          overall_box_a.bottom () + nh * box_height_a,
                          (nw == ntiles_w - 1) ? overall_box_a.right () : overall_box_a.left () + (nw + 1) * box_width_a,
                          (nh == ntiles_h - 1) ? overall_box_a.top () : overall_box_a.bottom () + (nh + 1) * box_height_a);

        db::Box region_b (overall_box_b.left () + nw * box_width_b, 
                          overall_box_b.bottom () + nh * box_height_b,
                          (nw == ntiles_w - 1) ? overall_box_b.right () : overall_box_b.left () + (nw + 1) * box_width_b,
                          (nh == ntiles_h - 1) ? overall_box_b.top () : overall_box_b.bottom () + (nh + 1) * box_height_b);

        region_a.enlarge (db::Point (tile_enlargement, tile_enlargement));
        region_b.enlarge (db::Point (tile_enlargement, tile_enlargement));

        for (std::vector <db::Coord>::iterator t = tolerances.begin (); t != tolerances.end (); ++t) {

          rdb::Category *subcat = 0;
          db::Cell *subcell = 0;

          if (output_mode == OMMarkerDatabase) {
            subcat = sub_categories[std::distance (tolerances.begin (), t)];
          } else {
            subcell = sub_cells[std::distance (tolerances.begin (), t)];
          }

          int layer_index = 0;
          for (std::map<db::LayerProperties, std::pair<int, int>, db::LPLogicalLessFunc>::const_iterator l = layers.begin (); l != layers.end (); ++l, ++layer_index) {

            ++progress;

            rdb::Category *layercat = 0;
            unsigned int layout_layer = 0;

            if (output_mode == OMMarkerDatabase) {
              layercat = layer_categories[std::distance (tolerances.begin (), t)][layer_index];
            } else {
              layout_layer = sub_output_layers[std::distance (tolerances.begin (), t)][layer_index];
            }

            int la = l->second.first;
            int lb = l->second.second;

            if ((la >= 0 && lb >= 0) || !summarize) {

              std::string desc;
              if (subcat) {
                desc = subcat->description () + " - " + tl::sprintf ("Layer %s, Tolerance %g, tile index=%d/%d,%d/%d", l->first.to_string ().c_str (), *t * dbu, nw + 1, ntiles_w, nh + 1, ntiles_h);
              } else {
                desc = tl::sprintf ("Layer %s, Tolerance %g, tile index=%d/%d,%d/%d", l->first.to_string ().c_str (), *t * dbu, nw + 1, ntiles_w, nh + 1, ntiles_h);
              }

              tl::SelfTimer timer (tl::verbosity () >= (has_tiles ? 21 : 11), desc);

              if (! xor_results.is_valid_layer (layer_index)) {

                xor_results.insert_layer (layer_index);

                if (la >= 0 && lb >= 0) {

                  if (! has_tiles) {

                    tl::SelfTimer timer (tl::verbosity () >= 21, "Boolean part");
#if 0
                    //  Straightforward implementation
                    sp.boolean (cva->layout (), cva->layout ().cell (cva.cell_index ()), la,
                                cvb->layout (), cvb->layout ().cell (cvb.cell_index ()), lb, 
                                xor_results_cell.shapes (layer_index), op, true, false, true);
#else
                    //  This implementation is faster when a lot of overlapping shapes are involved
                    db::Layout merge_helper;
                    db::Cell &merge_helper_cell = merge_helper.cell (merge_helper.add_cell ());
                    merge_helper.insert_layer (0);
                    merge_helper.insert_layer (1);

                    sp.merge (cva->layout (), cva->layout ().cell (cva.cell_index ()), la,
                              merge_helper_cell.shapes (0), true, 0, false, true);
                    sp.merge (cvb->layout (), cvb->layout ().cell (cvb.cell_index ()), lb,
                              merge_helper_cell.shapes (1), true, 0, false, true);
                    sp.boolean (merge_helper, merge_helper_cell, 0, 
                                merge_helper, merge_helper_cell, 1, 
                                xor_results_cell.shapes (layer_index), op, true, false, true);
#endif

                  } else {

                    tl::SelfTimer timer (tl::verbosity () >= 31, "Boolean part");
                    size_t n;

#if 0
                    //  Straightforward implementation
                    sp.clear ();

                    n = 0;
                    for (db::RecursiveShapeIterator s (cva->layout (), cva.cell (), la, region_a); ! s.at_end (); ++s, ++n) {
                      sp.insert (s.shape (), s.trans (), n * 2);
                    }

                    n = 0;
                    for (db::RecursiveShapeIterator s (cvb->layout (), cvb.cell (), lb, region_b); ! s.at_end (); ++s, ++n) {
                      sp.insert (s.shape (), s.trans (), n * 2 + 1);
                    }

                    db::BooleanOp bool_op (op);
                    db::ShapeGenerator sg (xor_results_cell.shapes (layer_index), true /*clear shapes*/);
                    db::PolygonGenerator out (sg, false /*don't resolve holes*/, false /*no min. coherence*/);
                    sp.process (out, op);
#else
                    //  This implementation is faster when a lot of overlapping shapes are involved
                    db::Layout merge_helper;
                    db::Cell &merge_helper_cell = merge_helper.cell (merge_helper.add_cell ());
                    merge_helper.insert_layer (0);
                    merge_helper.insert_layer (1);

                    //  This implementation is faster when a lot of overlapping shapes are involved
                    {
                      sp.clear ();

                      n = 0;
                      for (db::RecursiveShapeIterator s (cva->layout (), *cva.cell (), la, region_a); ! s.at_end (); ++s, ++n) {
                        sp.insert (s.shape (), s.trans (), n);
                      }

                      db::MergeOp op (0);
                      db::ShapeGenerator sg (merge_helper_cell.shapes (0), true /*clear shapes*/);
                      db::PolygonGenerator out (sg, false /*don't resolve holes*/, false /*no min. coherence*/);
                      sp.process (out, op);
                    }

                    {
                      sp.clear ();

                      n = 0;
                      for (db::RecursiveShapeIterator s (cvb->layout (), *cvb.cell (), lb, region_b); ! s.at_end (); ++s, ++n) {
                        sp.insert (s.shape (), s.trans (), n);
                      }

                      db::MergeOp op (0);
                      db::ShapeGenerator sg (merge_helper_cell.shapes (1), true /*clear shapes*/);
                      db::PolygonGenerator out (sg, false /*don't resolve holes*/, false /*no min. coherence*/);
                      sp.process (out, op);
                    }

                    sp.boolean (merge_helper, merge_helper_cell, 0, 
                                merge_helper, merge_helper_cell, 1, 
                                xor_results_cell.shapes (layer_index), op, true, false, true);
#endif

                  }

                } else if (op == db::BooleanOp::Xor || 
                           (op == db::BooleanOp::ANotB && la >= 0) ||
                           (op == db::BooleanOp::BNotA && lb >= 0)) {

                  db::RecursiveShapeIterator s;

                  if (la >= 0) {
                    if (has_tiles) {
                      s = db::RecursiveShapeIterator (cva->layout (), *cva.cell (), la, region_a);
                    } else {
                      s = db::RecursiveShapeIterator (cva->layout (), *cva.cell (), la);
                    }
                  } else if (lb >= 0) {
                    if (has_tiles) {
                      s = db::RecursiveShapeIterator (cvb->layout (), *cvb.cell (), lb, region_b);
                    } else {
                      s = db::RecursiveShapeIterator (cvb->layout (), *cvb.cell (), lb);
                    }
                  }

                  for (; ! s.at_end (); ++s) {
                    if (s->is_polygon () || s->is_box () || s->is_path ()) {
                      db::Polygon p;
                      s->polygon (p);
                      p.transform (s.trans ());
                      xor_results_cell.shapes (layer_index).insert (p);
                    }
                  }

                }

              }

              if (*t > 0) {
                tl::SelfTimer timer (tl::verbosity () >= 21, "Sizing part");
                sp.size (xor_results, xor_results_cell, layer_index, xor_results_cell.shapes (layer_index), -((*t + 1) / 2), (unsigned int)2, false);
                sp.size (xor_results, xor_results_cell, layer_index, xor_results_cell.shapes (layer_index), ((*t + 1) / 2), (unsigned int)2, false);
              }

              db::CplxTrans trans = db::CplxTrans (dbu);

              for (db::Shapes::shape_iterator s = xor_results_cell.shapes (layer_index).begin (db::ShapeIterator::All); ! s.at_end (); ++s) {

                if (has_tiles) {

                  std::vector <db::Polygon> clipped_poly;
                  clip_poly (s->polygon (), clip_box, clipped_poly);

                  for (std::vector <db::Polygon>::const_iterator cp = clipped_poly.begin (); cp != clipped_poly.end (); ++cp) {
                    if (output_mode == OMMarkerDatabase) {
                      rdb::Item *item = rdb->create_item (rdb_cell->id (), layercat->id ());
                      item->values ().add (new rdb::Value <db::DPolygon> (cp->transformed (trans)));
                    } else {
                      subcell->shapes (layout_layer).insert (*cp);
                    }
                  }
                
                } else {
                  if (output_mode == OMMarkerDatabase) {
                    rdb::Item *item = rdb->create_item (rdb_cell->id (), layercat->id ());
                    item->values ().add (new rdb::Value <db::DPolygon> (s->polygon ().transformed (trans)));
                  } else {
                    subcell->shapes (layout_layer).insert (s->polygon ());
                  }
                }

              }

            } else if (op == db::BooleanOp::Xor || 
                       (op == db::BooleanOp::ANotB && la >= 0) ||
                       (op == db::BooleanOp::BNotA && lb >= 0)) {

              if (la >= 0) {
                if (output_mode == OMMarkerDatabase) {
                  rdb::Item *item = rdb->create_item (rdb_cell->id (), layercat->id ());
                  item->values ().add (new rdb::Value <std::string> (tl::translate ("Layer not present at all in layout B")));
                }
              }

              if (lb >= 0) {
                if (output_mode == OMMarkerDatabase) {
                  rdb::Item *item = rdb->create_item (rdb_cell->id (), layercat->id ());
                  item->values ().add (new rdb::Value <std::string> (tl::translate ("Layer not present at all in layout A")));
                }
              }

            }
            
          }

        }

      }

    }

  }

  if (output_mode == OMMarkerDatabase) {
    mp_view->open_rdb_browser (rdb_index, cv_index_a);
  }

  mp_view->update_content ();

  if (output_mode == OMNewLayout && output_cell != 0 && output_cv >= 0) {
    mp_view->select_cell (output_cell->cell_index (), output_cv);
  }
}

}

